
#include <sys/stat.h>

#import "WiiDevControler.h"

@interface WiiDevControler(private) 
	- (void)checkDevice:(IOBluetoothDevice*)device;
	- (void)disconnectWiimote:(WiiRemote*)wiimote;
@end

static const int idleLEDs[] = {
	0x10, 0x20, 0x40, 0x90, 0x20, 0x40, 0x80, -1, 0,
};

static const NSTimeInterval idleDelays[] = {
	.05, .05, .05, .05, .05, .05, .05, .2, 1
};

#define NUM_IDLE_STATES 9

@implementation WiiRemote
- (void)reinit {
	[cchan writeSync:(unsigned char[]){0x52, 0x11, 0xF0} length:3];
	[cchan writeSync:(unsigned char[]){0x52, 0x12, 0x00, 0x30} length:4];
	[cchan writeSync:(unsigned char[]){0x52, 0x13, 0x00} length:3];
	[cchan writeSync:(unsigned char[]){0x52, 0x1A, 0x00} length:3];
}
@end

@implementation WiiDevControler

- (void)awakeFromNib {
	numWiiRemotes = 0;
	wiiRemotes = NULL;
	
	[statusLine setStringValue:@""];
	
	inquiry = nil;
	
	strcpy(basenameDisplay, "");
	snprintf(basename,104,"%s/Library/Wii Remotes/", getenv("HOME"));
	
	NSLog(@"%s", basename);
	
	mkdir(basename, 0755);
	// only do if needed and report error
	
	int i;
	for(i=0 ; i<16 ; i++) {
		struct stat ignored;
		char name[150];
		snprintf(name, 150, "%swii%d", basename, i);
		if (stat(name, &ignored) == 0) {
			unlink(name);
		}
		// readdir and report errors
	}
	
	idleState = 0;
	[self performSelector:@selector(idleAnim) withObject:nil afterDelay:idleDelays[idleState]];
	
	
	[[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(incomingConnection:)
		name:NSFileHandleConnectionAcceptedNotification object:nil];
	[[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(dataAvailable:)
		name:NSFileHandleDataAvailableNotification object:nil];
}

- (void)idleAnim {
	[self performSelector:@selector(idleAnim) withObject:nil afterDelay:idleDelays[idleState]];
	int i;
	unsigned char cmd[] = {0x52, 0x11, 0x00};
	cmd[2] = idleLEDs[idleState];
	
	for(i=0 ; i<numWiiRemotes ; i++) 
		if (wiiRemotes[i]->stream == -1) {
			if (idleLEDs[idleState] == -1) {
				cmd[2] = 0;
				if (wiiRemotes[i]->index & 1) cmd[2] |= 0x80;
				if (wiiRemotes[i]->index & 2) cmd[2] |= 0x40;
				if (wiiRemotes[i]->index & 4) cmd[2] |= 0x20;
				if (wiiRemotes[i]->index & 8) cmd[2] |= 0x10;
			}
			[wiiRemotes[i]->cchan writeSync:cmd length:3];
		}
	
	idleState++;
	idleState = idleState % NUM_IDLE_STATES;
}

- (NSApplicationTerminateReply)applicationShouldTerminate:(NSApplication *)sender {
	int i;
	
	for(i=0 ; i<numWiiRemotes ; i++) {
		[self disconnectWiimote:wiiRemotes[i]];
	}
	
	return NSTerminateNow;
}

- (void)disconnectWiimote:(WiiRemote*)wiimote {

	if (nil != wiimote->disconNote) {
		[wiimote->disconNote unregister];
		wiimote->disconNote = nil;
	}
	if (nil != wiimote->ichanNote) {
		[wiimote->ichanNote unregister];
		wiimote->ichanNote = nil;
	}
	if (nil != wiimote->cchanNote) {
		[wiimote->cchanNote unregister];
		wiimote->cchanNote = nil;
	}

	if (nil != wiimote->device) {
		if (nil != wiimote->cchan) {
			if ([wiimote->device isConnected])
				[wiimote->cchan closeChannel];
		//	[cchan release];
		}
		
		if (nil != wiimote->ichan) {
			if ([wiimote->device isConnected])
				[wiimote->ichan closeChannel];
		//	[ichan release];
		}
		
		[wiimote->device closeConnection];
	//	[wiiDevice release];
	}
	
	wiimote->ichan = wiimote->cchan = nil;
	wiimote->device = nil;
	
	close(wiimote->stream);
	close(wiimote->sock);
	unlink(wiimote->addr.sun_path);
	[wiimote release];
}

- (void)removeWiimote:(int)index {
	
	numWiiRemotes--;
	WiiRemote *wiimote = wiiRemotes[index];
	int i;
	for(i=index ; i<numWiiRemotes ; i++) {
		wiiRemotes[i] = wiiRemotes[i+1];
		wiiRemotes = realloc(wiiRemotes, sizeof(WiiRemote*)*numWiiRemotes);
	}
	[self disconnectWiimote:wiimote];
	
	[wiiList noteNumberOfRowsChanged];
}

- (void)windowWillClose:(NSNotification *)aNotification {
	
	[[NSApplication sharedApplication] terminate:self];
}

- (IBAction)disconect:(id)sender
{
	[self removeWiimote: [wiiList selectedRow]];
}

- (IBAction)sync:(id)sender
{
	inquiry = [IOBluetoothDeviceInquiry inquiryWithDelegate: self];
	
	if (nil == inquiry) {
		[statusLine setStringValue:@"Error: Failed to alloc IOBluetoothDeviceInquiry"];
	}
	
	[inquiry clearFoundDevices];
	
	[statusLine setStringValue:@"Preparing search..."];
	IOReturn ret = [inquiry start];
	
	if (ret == kIOReturnSuccess){
		[syncButton setEnabled:NO];
		[syncIndicator startAnimation:self];
		// copied from old code, is this necesary?  Shouldn't it already be retained once when it was initted?
		[inquiry retain];
	} else {
		[statusLine setStringValue:[NSString stringWithFormat:
			@"Error: Inquiry did not start, error %d", ret]];
		[inquiry setDelegate:nil];
		[inquiry release];
		inquiry = nil;
	}
}



// NSTableView Data Source methods

- (int)numberOfRowsInTableView:(NSTableView *)tableView {
	return numWiiRemotes;
}

- (id)tableView:(NSTableView *)tableView objectValueForTableColumn:(NSTableColumn *)tableColumn row:(int)row {
	if (row < 0 || row >= numWiiRemotes)
		return @"";
	
	WiiRemote *wiimote = wiiRemotes[row];
	
	switch([[tableColumn identifier] intValue]) {
		case 1:
			return wiimote->displayName;
		case 2:
			return (wiimote->stream != -1)?@"Yes":@"No";
		case 3:
			return [wiimote->device getAddressString];
		default:
			return @"Uknown column";
	}
}

- (void)tableViewSelectionDidChange:(NSNotification *)notification {
	if ([wiiList numberOfSelectedRows] == 0) {
		[disconnectButton setEnabled:NO];
	} else {
		[disconnectButton setEnabled:YES];
	}
}

/////// IOBluetoothDeviceInquiry delegates //////

- (void)	deviceInquiryStarted:(IOBluetoothDeviceInquiry*)sender {
	[statusLine setStringValue:@"Searching.  Press sync or 1 & 2 button."];
}

- (void) deviceInquiryComplete:(IOBluetoothDeviceInquiry*)sender 
        error:(IOReturn)error aborted:(BOOL)aborted {
	
	[inquiry autorelease];
	inquiry = nil;
	
	if (connecting == nil)
	[syncButton setEnabled:YES];
	[syncIndicator stopAnimation:self];
	
	if (kIOReturnSuccess != error) {
		[statusLine setStringValue:[NSString stringWithFormat:
			@"Error: Inquiry ended with error %d", error]];
	} else {
		[statusLine setStringValue:@"Search complete."];
	}
}

- (void)deviceInquiryDeviceNameUpdated:(IOBluetoothDeviceInquiry*)sender
		device:(IOBluetoothDevice*)device devicesRemaining:(int)devicesRemaining {
	[self checkDevice:device];
}

- (void) deviceInquiryDeviceFound:(IOBluetoothDeviceInquiry*)sender 
        device:(IOBluetoothDevice*)device {
	[self checkDevice:device];
}

/////// Connection Process /////////

- (int)indexForDevice:(IOBluetoothDevice*)device {
	int i;
	for(i=0 ; i<numWiiRemotes ; i++)
		if ([device isEqual:wiiRemotes[i]->device])
			return i;
	return -1;
}

- (void)checkDevice:(IOBluetoothDevice*)device {
//	if ([found objectForKey:[device getAddress]] == device)
//		return;
	if ([[device getName] isEqualToString:@"Nintendo RVL-CNT-01"]){
		int i = [self indexForDevice:device];
		if (i != -1) return;
		// new device!
		[inquiry stop];
		
		connecting = [[WiiRemote alloc] init];
		connecting->streamLock = [[NSLock alloc] init];
		connecting->deviceLock = [[NSLock alloc] init];
		connecting->device = device;
		IOReturn ret = [device openConnection:self];
		if (kIOReturnSuccess != ret) {
			[syncButton setEnabled:YES];
			[syncIndicator stopAnimation:self];
			
			[statusLine setStringValue:[NSString stringWithFormat:
				@"Error starting connection to Wii Remote (%08X)", ret]];
			return;
		} else {
			[statusLine setStringValue:@"Connecting..."];
		}
	}
}

- (void)connectionComplete:(IOBluetoothDevice*)device status:(IOReturn)ret {
	if (kIOReturnSuccess != ret) {
		[device closeConnection];
		[syncButton setEnabled:YES];
		[syncIndicator stopAnimation:self];
		
		[statusLine setStringValue:[NSString stringWithFormat:
			@"Error connectring to Wii Remote (%08X)", ret]];
		return;
	}
	
	connecting->device = device;
	
	connecting->disconNote = [device registerForDisconnectNotification:self
		selector:@selector(disconnected:fromDevice:)];
	
	// skipped SDP query - not needed and just slows down connection process
	
	connecting->cchan = connecting->ichan = nil;
	
	ret = [device openL2CAPChannelSync:&connecting->cchan withPSM:17 delegate:self];
	if (kIOReturnSuccess != ret) {
		[device closeConnection];
		[statusLine setStringValue:[NSString stringWithFormat:
			@"Error opening L2CAP Channel 17. (%08X)", ret]];
		NSLog(@"%s:%d", __FILE__, __LINE__);
		return;
	}
	[connecting->cchan retain];
	connecting->cchanNote = [connecting->cchan registerForChannelCloseNotification:self
		selector:@selector(channelClosed:channel:)];
	
	ret = [device openL2CAPChannelSync:&connecting->ichan withPSM:19 delegate:self];
	if (kIOReturnSuccess != ret) {
		[device closeConnection];
		[statusLine setStringValue:[NSString stringWithFormat:
			@"Error opening L2CAP Channel 19. (%08X)", ret]];
		NSLog(@"%s:%d", __FILE__, __LINE__);
		return;
	}
	[connecting->ichan retain];
	connecting->ichanNote = [connecting->ichan registerForChannelCloseNotification:self
		selector:@selector(channelClosed:channel:)];
	
	[connecting reinit];
	
	int index;
	
	for (index=1 ; ; index++) {
		int i;
		for(i=0 ; i<numWiiRemotes && wiiRemotes[i]->index != index ; i++) {}
		if (i == numWiiRemotes) break;
	}
	
	connecting->index = index;
	connecting->displayName = [[NSString alloc] initWithFormat:@"wii%d", index];
	
	connecting->addr.sun_family = AF_UNIX;
//	connecting->addr.sun_len = sizeof(connecting->addr);
	snprintf(connecting->addr.sun_path, 104, "%swii%d", basename, index);
	
/*	connecting->addr.sin_family = AF_INET;
	connecting->addr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
	connecting->addr.sin_port = htons(40000+index);*/
	
	connecting->sock = socket(AF_UNIX, SOCK_STREAM, 0);
	bind(connecting->sock, (void*)&connecting->addr, sizeof(connecting->addr));
	listen(connecting->sock, 0);
	
	connecting->stream = 0;
	
	[NSThread detachNewThreadSelector:@selector(serverThread:) toTarget:self withObject:connecting];
	
	numWiiRemotes++;
	wiiRemotes = realloc(wiiRemotes, sizeof(WiiRemote*)*numWiiRemotes);
	wiiRemotes[numWiiRemotes-1] = connecting;
	connecting = NULL;
	
	[wiiList noteNumberOfRowsChanged];
	
	[syncButton setEnabled:YES];
	[syncIndicator stopAnimation:self];
	[statusLine setStringValue:@""];
}

- (void)channelClosed: (IOBluetoothUserNotification*)note channel: (IOBluetoothL2CAPChannel*)channel  {
	int index = [self indexForDevice:[channel getDevice]];
	
	if (index >= 0) {
		WiiRemote *wiimote = wiiRemotes[index];
		[self removeWiimote:index];
		
		[statusLine setStringValue:[NSString stringWithFormat:
			@"Wii Remote %@ closed channel %d.", wiimote->displayName, [channel getRemoteChannelID]]];
		NSLog(@"%s:%d", __FILE__, __LINE__);
	} else {
		if (nil != connecting)
			[self disconnectWiimote:connecting];
		[statusLine setStringValue:[NSString stringWithFormat:
			@"Connecting Wii Remote closed channel %d.", [channel getRemoteChannelID]]];
		NSLog(@"%s:%d", __FILE__, __LINE__);
	}
}

- (void)disconnected: (IOBluetoothUserNotification*)note fromDevice: (IOBluetoothDevice*)device {
	int index = [self indexForDevice:device];
	
	if (index != -1) {
		WiiRemote *wiimote = wiiRemotes[index];
		[self removeWiimote:index];
			
		[statusLine setStringValue:[NSString stringWithFormat:
			@"Wii Remote %@ disconnected.", wiimote->displayName]];
		NSLog(@"%s:%d", __FILE__, __LINE__);
	} else {
		[statusLine setStringValue:@"Connecting Wii Remote  disconnected."];
		free(connecting);
		connecting = NULL;
		[syncButton setEnabled:YES];
		[syncIndicator stopAnimation:self];
	}
}

- (void)serverThread: (WiiRemote*)wiimote {
	[wiimote retain];
	
	while(nil != wiimote->device && [wiimote->device isConnected]) {
		struct sockaddr_un far;
		socklen_t farlen = sizeof(far);
		
		wiimote->stream = -1;
		
		[wiiList performSelectorOnMainThread:@selector(reloadData) withObject:nil waitUntilDone:NO];
		
		wiimote->stream = accept(wiimote->sock, (struct sockaddr*)&far, &farlen);
		
		[wiimote->streamLock lock];
		int value=1;
		setsockopt(wiimote->stream, SOL_SOCKET, SO_NOSIGPIPE, &value, sizeof(value));
		[wiimote->streamLock unlock];
		
		printf("accepted connection on fd %d\n", wiimote->stream);
		
		if (wiimote->stream < 0) break;
		
	/*	if (-1 == write(wiimote->stream, "test", 4))
			perror("initial write failed");
		else
			printf("initial write succeeded.\n");*/
		
		[wiiList performSelectorOnMainThread:@selector(reloadData) withObject:nil waitUntilDone:NO];
		
		unsigned char buf[256];
		int len;
		int stream = wiimote->stream;
		while(wiimote->stream != -1) {
			int readlen = recv(stream, buf, 1, MSG_WAITALL);
			if (readlen != 1) {
				printf("read packet length failed (%d)\n", readlen);
				break;
			}
			
			if (wiimote->stream != stream) break;
			
			len = buf[0];
			
			readlen = recv(stream, buf, len, MSG_WAITALL);
			if (readlen != len) {
				printf("read packet payload failed (%d != %d)\n", readlen, len);
				break;
			}
			
			switch (buf[0]) {
			case 0:
			
				break;
			case 0x52:
			//	printf("writing\n");
				[wiimote->cchan writeSync:buf length:len];
				break;
			default:
				break;
			}
		}
		[wiimote->streamLock lock];
		NSLog(@"%@ Exited read loop", wiimote->displayName);
		if (wiimote->stream != -1) {
			close(wiimote->stream);
			wiimote->stream = -1;
		}
		[wiimote->streamLock unlock];
		NSLog(@"%@ Stream closed", wiimote->displayName);
		
		[wiimote performSelectorOnMainThread:@selector(reinit) withObject:nil waitUntilDone:YES];
	}
	[wiimote release];
/*	if (cmd[0] == 0x52) {
		unsigned char *buf = calloc(1, length);
		memcpy(buf, cmd, length);
		printf("writing\n");
		[wiimote->cchan writeSync:buf length:length];
		free(buf);
	} else if (cmd[0] == 0) {
		unsigned char buf[16];
		buf[0] = 0;
		buf[1] = cmd[1];
		switch (cmd[1]) {
		case 0: // read MAC
			memcpy(buf+2, [wiimote->device getAddress]->data, 6);
			[sender writeData:[NSData dataWithBytes:buf length:8]];
		}
	}*/
}

-(void)l2capChannelData:(IOBluetoothL2CAPChannel*)l2capChannel data:(void *)dataPointer length:(size_t)length{
	IOBluetoothDevice *sender = [l2capChannel getDevice];
	
	int i;
	
	WiiRemote *wiimote = NULL;
	
	for(i=0 ; i<numWiiRemotes ; i++) {
		wiimote = wiiRemotes[i];
		if (wiimote->device == sender)
			break;
	}
	
	if (i == numWiiRemotes) {
		NSLog(@"Received data for unknown wiimote!");
		return;
	}
	
	
	[wiimote->streamLock lock];
	
	if (wiimote->stream != -1) {
		unsigned char header[2];
		header[0] = length+1;
		header[1] = [l2capChannel getLocalChannelID];
		
		// we care if these throw errors.
		
		int error = 0;
		
	//	printf("stream is %d\n", wiimote->stream);
		
		if (write(wiimote->stream, header, 2) != 2)
			perror(" write header failed"), error=1;
		/*else	
			printf(" write header succeeded\n");*/
		
		if (write(wiimote->stream, dataPointer, length) != length) 
			perror(" write payload failed"), error=1;
		/*else	
			printf(" write payload succeeded\n");*/
		if (error) {
			close(wiimote->stream);
			wiimote->stream = -1;
		}
	}
	[wiimote->streamLock unlock];
}

@end
